from reportlab.graphics.barcode import code39, code128, code93
from reportlab.graphics.barcode import eanbc, qr, usps, usps4s, common
from reportlab.graphics.shapes import Drawing
from reportlab.graphics import renderPDF

from pyametista.engine.commonutilities import replace_text, add_attr2attributes


def draw_basic_barcode(report, barcode_type, element, attributes):
    """
    Draw barcode on report_info.
    :param report: dictionary holding report information
    :param barcode_type: class + method to draw barcode depending of barcode type
    :param element: jrxml element list
    :param attributes: dictionary of attributes to use when drawing an object
    """
    code_expression = element.get('codeExpression')
    if code_expression is not None:
        data_value = str(replace_text(report, code_expression.get('value').strip('\"'), attributes))
        if attributes.get('checksumRequired', False):
            check_sum = 1
        else:
            check_sum = 0

        report['canvas'].saveState()

        barcode = barcode_type(data_value, barHeight=attributes.get('barHeight', attributes['height']), stop=1,
                       checksum=check_sum)

        # scale barcode to specified width
        report['canvas'].translate(attributes['x'], report['cur_y'] - attributes['y'] - attributes['height'])
        report['canvas'].scale(attributes.get('barWidth', attributes['width'])/barcode.width,
                             attributes.get('barHeight', attributes['height'])/ barcode.height)
        barcode.drawOn(report['canvas'], 0, 0)

        report['canvas'].restoreState()


def draw_advanced_barcode(report, barcode_type, element, attributes):
    """
    Advanced barcodes are barcodes that requires to be drawn to Drawing objects first.
    Barcodes that drawn to Drawing objects and then put on a report_info.
    :param report: dictionary holding report information
    :param barcode_type:
    :param element:
    :param attributes:
    """
    code_expression = element.get('codeExpression')
    if code_expression is not None:
        data_value = str(replace_text(report, code_expression.get('value').strip('\"'), attributes))
        barcode = barcode_type(data_value)
        bounds = barcode.getBounds()
        width = bounds[2] - bounds[0]
        height = bounds[3] - bounds[1]

        report['canvas'].saveState()
        d = Drawing(attributes['width'], attributes['height'],
                    transform=[attributes.get('barWidth', attributes['width'])/ width, 0, 0,
                               attributes.get('barHeight', attributes['height']) / height, 0, 0])
        d.add(barcode)
        renderPDF.draw(d, report['canvas'], attributes['x'], report['cur_y'] - attributes['y'] - attributes['height'])

        report['canvas'].restoreState()


"""
Values are (def to used to draw barcode on report_info, barcode component/def)
List of supported Barbecue barcodes. Supported list of Barcode4j barcodes must also be defined in
components.components.py
"""
barcode_types_dict = {
    'Codabar': (draw_basic_barcode, common.Codabar),
    'Code11': (draw_basic_barcode, common.Code11),
    'Code39': (draw_basic_barcode, code39.Standard39),
    'Code39 (Extended)': (draw_basic_barcode, code39.Extended39),
    'Code93': (draw_basic_barcode, code93.Standard93),
    'Code93 (Extended)': (draw_basic_barcode, code93.Extended93),
    'Code128': (draw_basic_barcode, code128.Code128),
    'EAN8': (draw_advanced_barcode, eanbc.Ean8BarcodeWidget),
    'EAN13': (draw_advanced_barcode, eanbc.Ean13BarcodeWidget),
    'EAN128': None,
    'FIM': (draw_basic_barcode, usps.FIM),   # TODO need to check
    'Interleaved2Of5': (draw_basic_barcode, common.I2of5),
    'Int2of5': (draw_basic_barcode, common.I2of5),
    'MSI': (draw_basic_barcode, common.MSI),   # TODO need to check
    'PostNet': (draw_basic_barcode, usps.POSTNET),
    'POSTNET': (draw_basic_barcode, usps.POSTNET),
    'QRCode': (draw_advanced_barcode, qr.QrCodeWidget),
    'USPS_4State': (draw_basic_barcode, usps4s.USPS_4State),
}


def process_barbecue(report, key, element, attributes):
    """
    Process jrxml barbecue component elements.
    :param report: dictionary holding report information
    :param element: jrxml barbecue element
    :param attributes: attributes of this element
    """
    add_attr2attributes(element, attributes)
    type = attributes.get('type')
    if type is not None:
        barcode_element_list = element.get('child')
        if barcode_element_list is not None:
            barcode_type = barcode_types_dict.get(type)
            if barcode_type is not None:
                barcode_type[0](report, barcode_type[1], barcode_element_list[0], attributes)


def process_barcode4j(report, key, element, attributes):
    """
    Process jrxml jarcode4j element.
    :param report: dictionary holding report information
    :param key:
    :param element:
    :param attributes:
    # :param row_data:  a row from data source
    """
    barcode_type = barcode_types_dict.get(key)

    if barcode_type is not None:
        barcode_type[0](report, barcode_type[1], element.get('child')[0], attributes)





from pyametista.adapters import CSVAdapter, DictAdapter, MysqlAdapter, PostgresqlAdapter
from pyametista.engine.bands.bands import (
    calc_column_footer_band_height,
    calc_page_footer_band_height,
    process_bands,
)
from pyametista.engine.bands.elements import process_jasper_report_element
from pyametista.engine.components.text import set_fonts
from pyametista.engine.jrxml2json import parse_jrxml
from pyametista.exports.pdf import create_canvas

try:
    import xml.etree.cElementTree as xml
except ImportError:
    import xml.etree.ElementTree as xml

import json
import logging
import logging.config
import os

logger = logging.getLogger(__name__)


def setup_logging(default_path="./logging.json", default_level=logging.INFO, env_key="LOG_CFG"):
    """
    Read logging configuration from file.
    :param default_path: default path to find logging configuration file
    :param default_level: default logging level
    :param env_key: read logging configuration default path from environmental variable
    :return:
    """
    path = default_path
    value = os.getenv(env_key, None)
    if value:
        path = value
    if os.path.exists(path):
        with open(path, "rt") as f:
            log_config = json.load(f)
        logging.config.dictConfig(log_config)
    else:
        logging.basicConfig(level=default_level)


def generate_report(jrxml_filename, output_filename, data_config, fonts=None, report_type="pdf"):
    """
    Generate pdf file.
    :param jrxml_filename: name of jrxml file to use.
    :param output_filename: name of output pdf file.
    :param data_config: data source configuration dictionary
    :param fonts: font configuration list
    :param report_type: type of file to generate (currently, only 'pdf' is supported)
    """
    setup_logging()

    if input is None or jrxml_filename is None:
        logger.error(
            "No report generated. Please specify jrxml filename and output filename to generate."
        )
    else:
        report_info = dict(
            properties={},
            variables={  # initial system variables
                "PAGE_NUMBER": {"value": 1, "class": "java.lang.Integer"},
                "MASTER_CURRENT_PAGE": {"value": None, "class": "java.lang.Integer"},
                "MASTER_TOTAL_PAGES": {"value": None, "class": "java.lang.Integer"},
                "COLUMN_NUMBER": {"value": 1, "class": "java.lang.Integer"},
                "REPORT_COUNT": {"value": 0, "class": "java.lang.Integer"},
                "PAGE_COUNT": {"value": 0, "class": "java.lang.Integer"},
                "COLUMN_COUNT": {"value": 0, "class": "java.lang.Integer"},
            },
            pre_row_data={},  # previous row from datasource
            row_data={},  # current row from datasource
        )
        report_info["main_datasource"] = None
        report_info["output_to_canvas"] = False
        report_info["group_names"] = {}
        report_info["group_cur"] = []
        report_info["prev_value"] = {}

        if "params" in data_config:
            data_params = data_config.pop("params")
            if data_params is not None:
                report_info["parameters"] = data_params

        if data_config is not None:
            data_adapter = data_config.pop("adapter")
            if data_adapter is None:
                logger.error(
                    "'adapter' should be specified in the data configuration:" + str(data_config)
                )
            else:
                if data_adapter == "mysql":
                    report_info["main_datasource"] = MysqlAdapter(**data_config)
                elif data_adapter == "postgres":
                    report_info["main_datasource"] = PostgresqlAdapter(data_config.get("config"))
                elif data_adapter == "csv":
                    report_info["main_datasource"] = CSVAdapter(data_config.get("filename"))
                elif data_adapter == "dict":
                    report_info["main_datasource"] = DictAdapter(data_config.get("data"))
                else:
                    logger.error(
                        "'invalid data adapter:"
                        + data_adapter
                        + "'. Valid adapter are 'mysql', 'postgres', 'csv'."
                    )
                    return
        try:
            jrxml_list = parse_jrxml(jrxml_filename)
            jasper_report_element = jrxml_list.get("jasperReport")
            process_jasper_report_element(report_info, jasper_report_element)

            # create report_info
            report_info["output_filename"] = output_filename
            # create_canvas(report_info)  # TODO move this to queryString. Note that there can be no queryString element

            # set fonts if any
            report_info["available_fonts"] = set_fonts(fonts)

            # set band's element to global report_info[]
            bands = jasper_report_element.get("child")
            for key, value in bands.items():
                report_info[key] = value

            # set column footer and footer heights. This is used to calculate available height for details band.
            report_info["col_footer_height"] = calc_column_footer_band_height(report_info)
            report_info["page_footer_height"] = calc_page_footer_band_height(report_info)
            report_info[
                "printingFooter"
            ] = False  # flag to denote if printing out column and page footers

            process_bands(report_info, bands)
        except FileNotFoundError as err:
            logging.error(
                'jrxml file "'
                + jrxml_filename
                + '" is not found. Please recheck the path and filename.'
            )


if __name__ == "__main__":
    filename = "only_group_header_with_field"
    # filename = 'group_without_expression'
    # filename = 'only_group_header'
    # filename = 'text_fields'

    input_filename = "../../tests/jrxml/" + filename + ".jrxml"  # input jrxml filename
    report_filename = "../../tests/output/pdf_" + filename + ".pdf"

    # MySQL datasource configuration
    config = {
        "adapter": "mysql",
        "host": "localhost",
        "user": "python",
        "password": "python",
        "database": "pyametista",
    }

    # Postgresql datasource configuration
    # config = {"adapter": "postgres",
    #                "config": "host='172.18.0.4' port='5432' dbname='pyametista' user='python' password='python'"}

    # CSV datasource configuration
    # config = {'adapter': 'csv', 'filename': '../../tests/data/address.csv'}

    font_list = [
        # list of additional directories to search for fonts
        {"font_path": ["../../tests/fonts/", "/usr/share/fonts/truetype/msttcorefonts/"]},
        # Japanese font
        {
            "font_filename": "ipag.ttc",
            "fonts": [{"index": 0, "name": "IPAGothic"}, {"index": 1, "name": "IPAPGothic"}],
        },
        # tests/fonts
        {"font_filename": "TIMES.TTF", "fonts": [{"index": 0, "name": "Times_New_Roman"}]},
        {"font_filename": "TIMESBD.TTF", "fonts": [{"index": 1, "name": "Times_New_Roman-Bold"}]},
        {"font_filename": "timesi.ttf", "fonts": [{"index": 2, "name": "Times_New_Roman-Italic"}]},
        {
            "font_filename": "TIMESBI0.TTF",
            "fonts": [{"index": 3, "name": "Times_New_Roman-BoldItalic"}],
        },
        {
            "font-family": {
                "name": "Times_New_Roman",
                "normal": "Times_New_Roman",
                "bold": "Times_New_Roman-Bold",
                "italic": "Times_New_Roman-Italic",
                "boldItalic": "Times_New_Roman-BoldItalic",
            }
        },
        # tests/fonts. No index
        {"font_filename": "Vera.ttf", "fonts": "Vera"},
        {"font_filename": "VeraBd.ttf", "fonts": "Vera-Bold"},
        {"font_filename": "VeraIt.ttf", "fonts": "Vera-Italic"},
        {"font_filename": "VeraBI.ttf", "fonts": "Vera-BoldItalic"},
        {
            "font-family": {
                "name": "Vera",
                "normal": "Vera",
                "bold": "Vera-Bold",
                "italic": "Vera-Italic",
                "boldItalic": "Vera-BoldItalic",
            }
        },
        # ubuntu font
        {"font_filename": "Verdana.ttf", "fonts": [{"index": 0, "name": "Verdana"}]},
        {"font_filename": "Verdana_Bold.ttf", "fonts": [{"index": 1, "name": "Verdana-Bold"}]},
        {"font_filename": "Verdana_Italic.ttf", "fonts": [{"index": 2, "name": "Verdana-Italic"}]},
        {
            "font_filename": "Verdana_Bold_Italic.ttf",
            "fonts": [{"index": 3, "name": "Verdana-BoldItalic"}],
        },
        {
            "font-family": {
                "name": "Verdana",
                "normal": "Verdana",
                "bold": "Verdana-Bold",
                "italic": "Verdana-Italic",
                "boldItalic": "Verdana-BoldItalic",
            }
        },
    ]

    generate_report(input_filename, report_filename, config, font_list)



DEFAULT_FONT='a'
DEFAULT_FONT_SIZE = 10

protocols = ('data://', 'file://', 'ftp://', 'http://', 'https://', 'mailto://', 'ssh://', 'tel://', 'urn://', 'ws://', 'wss://', 'view-source://')

workspace_prefix = ('file:')

db_field_type = {
 0: 'DECIMAL',
 1: 'TINY',
 2: 'SHORT',
 3: 'LONG',
 4: 'FLOAT',
 5: 'DOUBLE',
 6: 'NULL',
 7: 'TIMESTAMP',
 8: 'LONGLONG',
 9: 'INT24',
 10: 'DATE',
 11: 'TIME',
 12: 'DATETIME',
 13: 'YEAR',
 14: 'NEWDATE',
 15: 'VARCHAR',
 16: 'BIT',
 246: 'NEWDECIMAL',
 247: 'INTERVAL',
 248: 'SET',
 249: 'TINY_BLOB',
 250: 'MEDIUM_BLOB',
 251: 'LONG_BLOB',
 252: 'BLOB',
 253: 'VAR_STRING',
 254: 'STRING',
 255: 'GEOMETRY'
}


import logging

import psycopg2
from psycopg2.extras import DictCursor

logger = logging.getLogger(__name__)


class PostgresqlAdapter:
    """
    Postgresql datasource adapter.
    """

    # TODO support connection pooling
    # TODO implement bulk read

    def __init__(self, dsn):
        self.config = dsn
        try:
            self.conn = psycopg2.connect(self.config)
            # self.cur = self.conn.cursor()
            self.cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            # self.cur = self.conn.cursor(cursor_factory=psycopg2.extras.NamedTupleCursor)
        except psycopg2.OperationalError:
            self.close_cursor()
            self.conn = None
            self.cur = None
            logger.error("Operational error occurred connecting to Postgresql database.", exc_info=True)
        except psycopg2.Error:
            self.close_cursor()
            self.conn = None
            self.cur = None
            logger.error(
                "unable to connect to Postgresql database. internal database connection error.", exc_info=True
            )

    def get_cursor(self):
        """
        Return next row in query execution.
        execute_query() should be executed before calling this method.
        :return: next query result row.
        """
        return self.cur

    def execute_query(self, query_string):
        """
        Query a database table.
        :param query_string: query string to execute.
        :return: list of column names.
        """
        if self.cur is None:
            return None
        else:
            try:
                self.cur.execute(query_string)
                field_names = [i[0] for i in self.cur.description]
                return field_names
            except psycopg2.ProgrammingError as err:
                logging.error(
                    "Specified user does not have privilege to query the database table.", exc_info=True
                )
                return None
            except psycopg2.Error as err:
                logging.error("Unable to query the database table.", exc_info=True)

    def fetch_row(self):
        if self.cur is None:
            return None
        else:
            try:
                return self.cur.fetchone()
            except psycopg2.ProgrammingError as err:
                return None
            except psycopg2.Error as err:
                return None

    def close_cursor(self):
        """
        Close query execution cursor and connection.
        """
        try:
            self.cur.close()
            self.conn.close()
        except:
            pass


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    config = {
        "adapter": "postgres",
        "config": "host='172.18.0.4' port='5432' dbname='pyametista' user='python' password='python'",
    }
    sql = "SELECT * FROM orders"

    adapter = PostgresqlAdapter(config.get("config"))
    rows = adapter.execute_query(sql)

    logging.info(rows)
    adapter.close_cursor()



import logging

import mysql.connector
from mysql.connector import errorcode

logger = logging.getLogger(__name__)


class MysqlAdapter:
    """
    MySQL datasource adapter.
    """

    # TODO support connection pooling
    # TODO implement bulk read

    def __init__(self, **config_list):
        self.config = config_list
        try:
            self.conn = mysql.connector.connect(**config_list)
            self.cur = self.conn.cursor()
        except mysql.connector.Error as err:
            self.close_cursor()
            self.conn = None
            self.cur = None
            if err.errno == errorcode.ER_ACCESS_DENIED_ERROR:
                logger.error("Unable to access MySQL database with provided user/password.")
            elif err.errno == errorcode.ER_BAD_DB_ERROR:
                logger.error("Specified MySQL database does not exist")
            else:
                logger.error("Failed to connect to MySQL database", exc_info=True)
        except AttributeError as err:
            self.close_cursor()
            self.conn = None
            self.cur = None
            print("invalid connection configuration:" + str(self.config))

    def get_cursor(self):
        """
        Return next row in query execution.
        execute_query() should be executed before calling this method.
        :return: next query result row.
        """
        return self.cur

    def execute_query(self, query_string):
        """
        Query a database table.
        :param query_string: query string to execute.
        :return: list of column names.
        """
        if self.cur is None:
            return None
        else:
            self.cur.execute(query_string)
            field_names = [i[0] for i in self.cur.description]
            return field_names

    def fetch_row(self):
        if self.cur is None:
            return None
        else:
            return self.cur.fetchone()

    def close_cursor(self):
        """
        Close query execution cursor and connection.
        """
        try:
            self.cur.close()
            self.conn.close()
        except Exception:
            pass


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    config = {
        "host": "localhost",
        "port": "3306",
        "user": "python",
        "password": "python",
        "database": "pyametista",
    }
    sql = "SELECT * FROM orders"

    adapter = MysqlAdapter(**config)
    rows = adapter.execute_query(sql)
    logger.info(rows)
    adapter.close_cursor()



from pyametista.engine.commonutilities import add_attr2attributes
from reportlab.lib.pagesizes import A4


def process_report_element(report, element):
    """
    Set attributes of "reportElement" to a dictionary.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    :return: dictionary of attributes of specified element
    """
    text_settings = element.get("attr")
    attributes = {
        "x": int(text_settings.get("x")) + report["leftMargin"],
        "y": int(text_settings.get("y")),
        "height": int(text_settings.get("height", 0)),
        "width": int(text_settings.get("width", 0)),
        "forecolor": text_settings.get("forecolor"),
        "backcolor": text_settings.get("backcolor"),
        "mode": text_settings.get("mode"),  # 'Opaque': show
    }
    return attributes


def process_graphic_element(report, element, attributes):
    """
    Process jrxml graphicElement.
    Get attributes on graphicElement.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    :param attributes: current dictionary of attributes.
    :return: attributes with attributes of pen under graphicElements.
    """
    if len(element) > 1:
        graphic_element = element[1].get("graphicElement")  # get graphicElement
        if graphic_element is not None:
            pen = graphic_element.get("child")
            if pen is not None:
                add_attr2attributes(pen[0].get("pen"), attributes)
    return attributes


def process_jasper_report_element(report, element):
    """
    Get page attributes from the attributes in jrxml 'jasperReport' element.
    Set obtained values to global 'report_info' variable. Default values are defined here.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    """
    default_page_width, default_page_height = A4  # set default page size to A4

    page_settings = element.get("attr")

    report["pageWidth"] = float(page_settings.get("pageWidth", default_page_width))
    report["pageHeight"] = float(page_settings.get("pageHeight", default_page_height))
    report["name"] = page_settings.get("name", "")
    report["language"] = page_settings.get("language", "python")
    report["columnCount"] = float(page_settings.get("columnCount", "1"))
    report["printOrder"] = page_settings.get("printOrder", "Vertical")
    report["orientation"] = page_settings.get("orientation", "Portrait")
    report["whenNoDataType"] = page_settings.get("whenNoDataType", "Null")
    report["columnWidth"] = float(page_settings.get("columnWidth", "555"))
    report["columnSpacing"] = float(page_settings.get("columnSpacing", "0"))
    report["leftMargin"] = float(page_settings.get("leftMargin", "20"))
    report["rightMargin"] = float(page_settings.get("rightMargin", "20"))
    report["topMargin"] = float(page_settings.get("topMargin", "30"))
    report["bottomMargin"] = float(page_settings.get("bottomMargin", "30"))
    report["isTitleNewPage"] = page_settings.get("isTitleNewPage", "false")
    report["isSummaryNewPage"] = page_settings.get("isSummaryNewPage", "false")
    report["isFloatColumnFooter"] = page_settings.get("isFloatColumnFooter", "false")
    report["scriptletClass"] = page_settings.get("scriptletClass", "")
    report["resourceBundle"] = page_settings.get("resourceBundle", "")
    report["whenResourceMissingType"] = page_settings.get("whenResourceMissingType", "Null")
    report["isIgnorePagination"] = page_settings.get("isIgnorePagination", "false")
    report["formatFactoryClass"] = page_settings.get("formatFactoryClass", "")

    # current y axis. To force a new page initially, add +1 to page height
    # report['cur_y'] = report['page_height'] + 1
    report["cur_y"] = report["pageHeight"] - report["topMargin"]



import logging

logger = logging.getLogger(__name__)


class DictAdapter:
    """
    Dict adapter.
    """

    def __init__(self, data):
        self.data = iter(data)

    def get_cursor(self):
        """
        Return next row in data
        execute_query() should be executed before calling this method.
        :return: next data row.
        """
        return next(self.data)
        # return None

    def execute_query(self, query_string):
        """
        Query a database table.
        :param query_string: query string to execute.
        :return: list of column names.
        """
        if self.data is None:
            return None
        else:
            return next(self.data)

    def fetch_row(self):
        if self.data is None:
            return None
        else:
            try:
                return next(self.data)
            except StopIteration as err:
                return None

    def close_cursor(self):
        """
        Close query execution cursor and connection.
        """
        del self.data


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    data_config = {"adapter": "dict", "data": [{"test_row": "1", "testing": "2"}]}

    adapter = DictAdapter(data_config.get("data"))
    header = adapter.execute_query("")
    logger.info(header)
    adapter.close_cursor()



import datetime
import logging

from reportlab import rl_config
from reportlab.lib import colors
from reportlab.lib.enums import TA_CENTER, TA_JUSTIFY, TA_LEFT, TA_RIGHT
from reportlab.lib.styles import ParagraphStyle, getSampleStyleSheet
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.platypus import KeepInFrame, Paragraph, Table, TableStyle

from pyametista.engine.bands.elements import process_report_element
from pyametista.engine.commonutilities import add_attr2attributes, convert2boolean, replace_text
from pyametista.engine.components.line import process_box_element

logger = logging.getLogger(__name__)


def set_fonts(fonts_list):
    """
    Register specified fonts and font families to report_info.
    :param fonts_list: list of fonts and font families to register
    """
    # TODO scan jrxml and try to automatically load fonts

    available_fonts = []
    if fonts_list is not None:
        for font in fonts_list:
            font_path = font.get("font_path")
            if type(font_path) == list:
                rl_config.TTFSearchPath.extend(font_path)
            font_filename = font.get("font_filename")
            font_list = font.get("fonts")
            if font_filename is not None and font_list is not None:
                if type(font_list) is str:
                    try:
                        pdfmetrics.registerFont(TTFont(font_list, font_filename))
                        available_fonts.append(font_list)
                    except:  # if font files is not found, output error and skip it.
                        logger.error(
                            "font not found. font name:" + name + " file name:" + font_filename
                        )
                elif type(font_list) is list:
                    for font_reg in font_list:
                        index = font_reg.get("index")
                        name = font_reg.get("name")
                        if name is not None:
                            if index is not None:
                                try:
                                    pdfmetrics.registerFont(
                                        TTFont(name, font_filename, subfontIndex=index)
                                    )
                                    available_fonts.append(name)
                                except:  # if font files is not found, output error and skip it.
                                    logger.error(
                                        "font not found. font name:"
                                        + name
                                        + " file name:"
                                        + font_filename
                                    )
                            else:
                                try:
                                    pdfmetrics.registerFont(TTFont(name, font_filename))
                                    available_fonts.append(name)
                                except:  # if font files is not found, output error and skip it.
                                    logger.error(
                                        "font not found. font name:"
                                        + name
                                        + " file name:"
                                        + font_filename
                                    )
            font_family = font.get("font-family")
            if font_family is not None:
                if (
                    font_family.get("name") is not None
                    and font_family.get("normal") is not None
                    and font_family.get("bold") is not None
                    and font_family.get("italic") is not None
                    and font_family.get("boldItalic") is not None
                ):
                    pdfmetrics.registerFontFamily(
                        font_family.get("name"),
                        normal=font_family.get("normal"),
                        bold=font_family.get("bold"),
                        italic=font_family.get("italic"),
                        boldItalic=font_family.get("boldItalic"),
                    )
    return available_fonts


def format_text(text, attributes):
    """
    Format text according to specified pattern.
    :param text: text to format
    :param attributes: attributes corresponding to the text element
    :return: formatted text according to pattern
    """

    if attributes.get("pattern") is not None:
        if type(text) is str:
            pass  # string do not have pattern
        elif type(text) is int:
            if "{:" in attributes["pattern"]:
                text = attributes["pattern"].format(int(text))
        elif type(text) is float:
            if "{:" in attributes["pattern"]:
                text = attributes["pattern"].format(int(text))
        elif type(text) is complex:
            if "{:" in attributes["pattern"]:
                text = attributes["pattern"].format(int(text))
        elif type(text) is bool:
            pass
        elif isinstance(text, datetime.datetime):
            if "%" in attributes.get("pattern"):
                text = text.strftime(attributes.get("pattern"))
    return str(text)


def process_font(element, attributes):
    """
    Process jrxml 'font' element.
    Set font attributes.
    :param element: jrxml 'font' element
    :param attributes: attributes related to the 'element'
    """
    add_attr2attributes(element, attributes, "font")


def process_paragraph(element, attributes):
    """
    Process jrxml 'paragraph' element.
    :param element: jrxml 'paragraph' element
    :param attributes: attributes related to the 'element'
    """
    add_attr2attributes(element, attributes)


"""Elements within text element."""
textElement_dict = {"font": process_font, "paragraph": process_paragraph}


def process_text_element(report, element, attributes):
    """
    textElement is used to specify alignment of the following text
    :param report: dictionary holding report information
    :param element:
    :param attributes:
    :return:
    """
    textElement = element.get("child")
    if textElement is not None:
        for tag in textElement:
            for key, value in tag.items():
                if textElement_dict[key] is not None:
                    textElement_dict[key](value, attributes)

    text_settings = element.get("attr")
    if text_settings is not None:
        for key, value in text_settings.items():
            attributes[key] = value
    return attributes


def process_text(report, element, attributes):
    """
    Process "text" jrxml element.
    :param report: dictionary holding report information
    :param element:
    :param attributes:
    """
    draw_text(report, element.get("value", ""), attributes)


static_text_dict = {
    # 'reportElement': process_report_element,
    "box": process_box_element,
    "textElement": process_text_element,
    "text": process_text,
}


def process_static_text(report, element):
    """
    Process jrxml staticText element.
    :param report: dictionary holding report information
    :param element:
    :param row_data: a row from data source
    """
    static_text_element = element.get("child")
    if static_text_element is not None:
        report_element = process_report_element(
            report, static_text_element[0].get("reportElement")
        )  # get reportElement
        for tag in static_text_element[1:]:
            for key, value in tag.items():
                if static_text_dict[key] is not None:
                    if key == "textElement":
                        report_element = process_text_element(report, value, report_element)
                    else:
                        static_text_dict[key](
                            report=report, element=value, attributes=report_element
                        )


def text_element(element):
    # process_text(element)
    pass


def process_text_fieldExpression(report, element, attributes):
    """
    Process jrxml textFieldExpression element.
    :param report: dictionary holding report information
    :param element:
    :param attributes:
    """
    data_value = replace_text(report, element.get("value"), attributes)
    # if report['group_cur']:     # check if this is being processed within a group
    #     group_name = report['group_cur'][-1]
    #     # print(group_name, report['group_names'])
    #     prev_value = report['group_names'].get(group_name)
    #     # print(prev_value, data_value)
    #     if prev_value is None or data_value != prev_value:
    #         # report['group_names'].update({group_name: data_value})
    #         draw_text(report, data_value, attributes)
    # else:   # ['group_cur'] is empty -> not in a group then output value
    draw_text(report, data_value, attributes)


"""
possible child elements under textField element
"""
textField_dict = {
    "reportElement": process_report_element,
    "box": process_box_element,
    "textElement": process_text_element,
    "textFieldExpression": process_text_fieldExpression,
}


def process_text_field(report, element):
    """
    Process textField jrxml element.
    :param report: dictionary holding report information
    :param element:
    # :param row_data: a row from data source
    """
    text_field = element.get("child")
    if text_field is not None:
        attributes = process_report_element(
            report, text_field[0].get("reportElement")
        )  # get reportElement

        # set attributes to reportElement (e.g. pattern, isBlankWhenNull)
        add_attr2attributes(element, attributes)

        for tag in text_field[1:]:
            for key, value in tag.items():
                if textField_dict[key] is not None:
                    if key == "textElement":
                        attributes = process_text_element(report, value, attributes)
                    else:  # reportElement, box
                        textField_dict[key](report, value, attributes)


def get_font(report, report_element, base_style):
    """
    Get specified font name.
    Default to base style font (Usually 'helvetica') if the font is not available in the report_info.
    :param report: dictionary holding report information
    :param report_element: jrxml reportElement element
    :param base_style: report_info base style
    :return: valid font name
    """
    font_name = report_element.get("fontFontName", base_style.fontName)
    if font_name is not None:
        font_name.strip()
        if (
            font_name not in report["canvas"].getAvailableFonts()
            and font_name not in report["available_fonts"]
        ):
            logging.warn(
                'font file "'
                + font_name
                + '" not found. Replacing with font "'
                + base_style.fontName
                + '"'
            )
            font_name = base_style.fontName
    return font_name


def draw_text(report, text, attributes):
    """
    Draw text string on report_info.
    :param report:
    :param text: text to output to report_info
    :param attributes: attributes (e.g. 'font name', 'font size', 'color') to apply to text
    """
    report["canvas"].saveState()

    text = format_text(text, attributes)
    styles = getSampleStyleSheet()

    text_alignment = {
        "Left": TA_LEFT,
        "Center": TA_CENTER,
        "Right": TA_RIGHT,
        "Justified": TA_JUSTIFY,
    }
    base_style = styles["Normal"]
    left_indent = attributes.get("leftIndent", base_style.leftIndent)
    right_indent = attributes.get("rightIndent", base_style.leftIndent)
    font_name = get_font(report, attributes, base_style)

    font_is_bold = convert2boolean(attributes.get("fontIsBold"))
    if font_is_bold:
        text = "<b>" + text + "</b>"
    font_is_italic = convert2boolean(attributes.get("fontIsItalic"))
    if font_is_italic:
        text = "<i>" + text + "</i>"

    font_is_underline = convert2boolean(attributes.get("fontIsUnderline"))
    if font_is_underline:
        text = "<u>" + text + "</u>"
    font_is_strike_through = convert2boolean(attributes.get("fontIsStrikeThrough"))
    if font_is_strike_through:
        text = "<strike>" + text + "</strike>"

    font_size = attributes.get("fontSize", 10)
    text_color = attributes.get("forecolor")
    if text_color is None:
        text_color = base_style.textColor
    else:
        text_color = colors.HexColor(text_color)

    ps = ParagraphStyle(
        name="cell",
        parent=styles["Normal"],
        fontName=font_name,
        fontSize=font_size,
        leading=font_size * 1.2,
        leftIndent=left_indent,
        rightIndent=right_indent,
        alignment=text_alignment[attributes.get("textAlignment", "Left")],
        textColor=text_color,
    )
    # text = "<br/>".join(str(text).split('\n'))   # replace '\n' with a <br/> to cause newline
    text = str(text).replace("\n", "<br/>")
    # story = [Paragraph(str(text), ps)]
    story = [Paragraph(text, ps)]
    story_in_frame = KeepInFrame(
        attributes["width"], attributes["height"], story, mode="shrink"
    )  # 'truncate, overflow, shrink

    data = [[story_in_frame]]
    t = Table(data, colWidths=attributes["width"], rowHeights=attributes["height"])

    table_style = TableStyle(
        [
            ("VALIGN", (0, 0), (0, 0), attributes.get("verticalAlignment", "TOP").upper()),
            ("TEXTCOLOR", (0, 0), (0, 0), colors.black),
            ("LEFTPADDING", (0, 0), (0, 0), 0),
            ("RIGHTPADDING", (0, 0), (0, 0), 0),
            ("TOPPADDING", (0, 0), (0, 0), 0),
            ("BOTTOMPADDING", (0, 0), (0, 0), 0),
        ]
    )
    if attributes.get("mode") is not None:
        backcolor = attributes.get("backcolor")
        if backcolor is not None:
            table_style.add("BACKGROUND", (0, 0), (0, 0), colors.HexColor(backcolor))

    t.setStyle(table_style)
    t.wrapOn(report["canvas"], attributes["width"], attributes["height"])
    t.drawOn(
        report["canvas"], attributes["x"], report["cur_y"] - attributes["y"] - attributes["height"]
    )

    report["canvas"].restoreState()


def process_text(element, attributes):
    """
    Process "text" jrxml element.
    :param element:
    :param attributes:
    """
    draw_text(element.get("value", ""), attributes)



"""
Jrxml2json.py converts jrxml to json format.
"""
try:
    import xml.etree.cElementTree as xml
except ImportError:
    import xml.etree.ElementTree as xml

import json

import logging
logger = logging.getLogger(__name__)


def strip_namespace(element):
    """
    strip namespace from the element name.
    :param element: etree element.
    :return: element name without a namespace.
    """
    return element.tag.split('}')[-1]


def get_child_dict(element):
    """
    recursively create a tuple of child, attribute, and value of the element.
    :param element: etree element
    :return: an list with tuple of child, attribute, and value of the element
    """
    child_list = list(append_child_elements(element))
    if len(child_list) > 0:
        yield ('child', child_list)
    if len(element.attrib) > 0:
        yield ('attr', element.attrib)
    if element.text is not None and len(element.text.strip()) > 0:
        yield ('value', element.text)


def append_child_elements(element):
    """
    recursively append dict of element and it's child, attribute, and value of the element.
    :param element: etree element
    :return: dictionary of element name and sttributes, child elements, and value
    """
    for child in element:
        yield dict([(strip_namespace(child), dict(get_child_dict(child)))])


def parse_jrxml(filename):
    """
    Parse input jrxml file and convert it to json format.
    :param filename: jrxml file to parse
    :return: content of jrxml file in json format
    """
    xml_doc = xml.parse(filename)
    root = xml_doc.getroot()

    band_dict = {}
    for band in root:
        band_name = strip_namespace(band)
        band_element = band_dict.get(band_name)
        if band_element is None:
            band_dict[band_name] = list([{band_name: dict(get_child_dict(band))}])
        else:
            band_element.extend(list([{band_name: dict(get_child_dict(band))}]))
        #     logger.debug(band_element)
            # band_dict.extend(list([{band_name: dict(get_child_dict(band))}]))

    return dict([(strip_namespace(root), {'attr': root.attrib, 'child': band_dict})])


def write2file(element, json_filename):
    """
    Write json to a file
    :param element: json element to write to a file
    :param json_filename: name of the output file
    :return: None
    """
    try:
        with open(json_filename,"w") as out_file:
            out_file.write(str(element))
    except IOError as err:
        logger.error(err)


if __name__ == '__main__':
    filename = 'empty_report'
    input_filename = '../../test/jrxml/' + filename + '.jrxml'  # input jrxml filename
    output_filename = '../../test/output/' + filename + '.json'

    parsed_jrxml = parse_jrxml(input_filename)
    write2file(json.dumps(parsed_jrxml), output_filename)
    # logger.debug(json.dumps(parsed_jrxml, indent=2))






from pyametista.adapters.CSVAdapter import CSVAdapter
from pyametista.adapters.DictAdapter import DictAdapter
from pyametista.adapters.MysqlAdapter import MysqlAdapter
from pyametista.adapters.PostgresqlAdapter import PostgresqlAdapter



from pyametista.engine.components.barcode import process_barbecue, process_barcode4j
from pyametista.engine.bands.elements import process_report_element

"""
Supported components. Barcode4j barcodes are also defined here.
"""
componentElement_dict = {
    "reportElement": None,
    "barbecue": process_barbecue,
    "Codabar": process_barcode4j,
    "Code11": process_barcode4j,
    "Code128": process_barcode4j,
    "Code39": process_barcode4j,
    "Code39 (Extended)": process_barcode4j,
    "Code93": process_barcode4j,
    "Code93 (Extended)": process_barcode4j,
    "EAN8": process_barcode4j,
    "EAN13": process_barcode4j,
    "EAN128": process_barcode4j,
    "FIM": process_barcode4j,
    "Interleaved2Of5": process_barcode4j,
    "MSI": process_barcode4j,
    "PostNet": process_barcode4j,
    "POSTNET": process_barcode4j,
    "QRCode": process_barcode4j,
    "USPS_4State": process_barcode4j,
}


def process_componentElement(report, element):
    """
    Process component jrxml element.
    :param report: dictionary holding report information
    :param element: jrxml component element
    # :param row_data: a row from data source
    """
    component_element = element.get("child")
    if component_element is not None:
        report_element = process_report_element(
            report, component_element[0].get("reportElement")
        )  # get reportElement
        for tag in component_element[1:]:
            for key, value in tag.items():
                if componentElement_dict.get(key) is not None:
                    componentElement_dict[key](report, key, value, report_element)



import os

from pyametista.engine import reportGenerator


class BasicReport:
    """ "
    Base wrapper class to generate reports.
    """

    def __init__(
        self, jrxml_filename, output_filename, data_config=None, fonts=None, report_type="pdf"
    ):
        self.jrxml_filename = jrxml_filename
        self.output_filename = output_filename
        self.data_config = data_config

        if fonts is not None:
            self.fonts = fonts
        else:
            self.fonts = [
                {"font_path": [os.path.join(os.path.abspath(os.path.dirname(__file__)), "fonts")]},
                {
                    "font_filename": "OpenSans-Regular.ttf",
                    "fonts": [{"index": 0, "name": "OpenSans"}],
                },
                {
                    "font_filename": "OpenSans-Bold.ttf",
                    "fonts": [{"index": 1, "name": "OpenSans-Bold"}],
                },
                {
                    "font_filename": "OpenSans-Italic.ttf",
                    "fonts": [{"index": 2, "name": "OpenSans-Italic"}],
                },
                {
                    "font_filename": "OpenSans-BoldItalic.ttf",
                    "fonts": [{"index": 3, "name": "OpenSans-BoldItalic"}],
                },
                {
                    "font-family": {
                        "name": "OpenSans",
                        "normal": "OpenSans",
                        "bold": "OpenSans-Bold",
                        "italic": "OpenSans-Italic",
                        "boldItalic": "OpenSans-BoldItalic",
                    }
                },
                {"font_filename": "Roboto-Regular.ttf", "fonts": [{"index": 0, "name": "Roboto"}]},
                {
                    "font_filename": "Roboto-Bold.ttf",
                    "fonts": [{"index": 1, "name": "Roboto-Bold"}],
                },
                {
                    "font_filename": "Roboto-Italic.ttf",
                    "fonts": [{"index": 2, "name": "Roboto-Italic"}],
                },
                {
                    "font_filename": "Roboto-BoldItalic.ttf",
                    "fonts": [{"index": 3, "name": "Roboto-BoldItalic"}],
                },
                {
                    "font-family": {
                        "name": "Roboto",
                        "normal": "Roboto",
                        "bold": "Roboto-Bold",
                        "italic": "Roboto-Italic",
                        "boldItalic": "Roboto-BoldItalic",
                    }
                },
                {
                    "font_filename": "RobotoSlab-Regular.ttf",
                    "fonts": [{"index": 0, "name": "RobotoSlab"}],
                },
                {
                    "font_filename": "RobotoSlab-Bold.ttf",
                    "fonts": [{"index": 1, "name": "RobotoSlab-Bold"}],
                },
                {
                    "font-family": {
                        "name": "RobotoSlab",
                        "normal": "RobotoSlab",
                        "bold": "RobotoSlab-Bold",
                    }
                },
            ]

        self.report_type = report_type

    def generate_report(self):
        reportGenerator.generate_report(
            jrxml_filename=self.jrxml_filename,
            output_filename=self.output_filename,
            data_config=self.data_config,
            fonts=self.fonts,
            report_type=self.report_type,
        )

    def set_report_type(self, report_type):
        self.report_type = report_type



import datetime
import re


def is_bytes(data):
    """
    Check if data is of type byte or bytearray.
    :param data:
    :return:
    """
    try:
        data = data.decode()
        return False
    except AttributeError:
        return True


def add_attr2attributes(element, attributes, prefix=None):
    """
    Add attributes of elements to 'attributes' dictionary.
    :param element:
    :param attributes:
    :param prefix:
    :return:
    """
    attr = element.get("attr")
    if attr is not None:
        for key, value in attr.items():
            if prefix is None:
                dict_key = key
            else:
                dict_key = prefix + key[0].upper() + key[1:]
            if value.isdigit():
                attributes[dict_key] = int(value)
            else:
                attributes[dict_key] = value
    # return attributes


# def replace_fields(field_dict, expression, row_data, attributes):
def replace_fields(report, expression, attributes):
    """
    Replace fields with value from data source.
    If the field is not in the database, do not replace.
    :param report:
    :param expression: text to be processed
    :param attributes: attribute of the element being processed (i.e. 'isBlankWhenNull')
    :return: expression with $F{} replaced with values from data source when they exist
    """
    if expression is None or report["row_data"] is None:
        return None
    else:
        # find all field keys in specified expression
        new_keys = [key[key.find("$F{") + 3 : key.find("}")] for key in re.findall(r"\$F\{.*?\}", expression)]
        if report.get("field_dict") is not None:
            for key in new_keys:
                data_type = report["field_dict"].get(key)
                if data_type == "java.lang.Integer":
                    data_value = str(report["row_data"].get(key, "$F{" + key + "}"))
                elif data_type == "java.sql.Timestamp":
                    date_time = report["row_data"].get(key, "$F{" + key + "}")
                    if type(date_time) is not datetime.datetime:
                        data_value = date_time
                    else:
                        data_value = (
                            "datetime.datetime("
                            + str(date_time.year)
                            + ","
                            + str(date_time.month)
                            + ","
                            + str(date_time.day)
                            + ","
                            + str(date_time.hour)
                            + ","
                            + str(date_time.minute)
                            + ","
                            + str(date_time.second)
                            + ","
                            + str(date_time.microsecond)
                            + ","
                            + str(date_time.tzinfo)
                            + ")"
                        )
                elif data_type == "java.lang.Object":
                    return report["row_data"].get(key, "$F{" + key + "}")
                else:
                    is_blank_when_null = convert2boolean(attributes.get("isBlankWhenNull"))
                    value = report["row_data"].get(key, "$F{" + key + "}")
                    if is_blank_when_null and value is None:
                        # data_value = ''
                        return ""
                    else:
                        data_value = '"' + str(report["row_data"].get(key, "$F{" + key + "}")) + '"'
                if key is not None and data_value is not None:
                    expression = expression.replace("$F{" + key + "}", data_value)
        return expression


def replace_variables(variables, expression, attributes):
    """
    Replace variables with value from data source.
    If the variable is not in the database, do not replace.
    :param variables:
    :param expression:
    :param attributes:
    :return:
    """
    return replace_data("V", variables, expression, attributes)


def replace_parameters(parameters, expression, attributes):
    """
    Replace parameters with value from data source.
    If the variable is not in the database, do not replace.
    :param parameters:
    :param expression:
    :param attributes:
    :return:
    """
    return replace_data("P", parameters, expression, attributes)


def replace_data(type: str, variables, expression, attributes):
    """
    Replace field/variables/parameter with value from respective source.
    If the variable is not found, do not replace.
    :param type:
    :param variables:
    :param expression:
    :param attributes:
    :return:
    """

    if type != "V" and type != "P":
        raise ValueError("Invalid type! Must be 'V' or 'P'!")

    regex_str = r"\$" + type + r"\{.*?\}"

    # Find all keys in specified expression
    new_keys = [key[key.find(f"${type}{{") + 3 : key.find("}")] for key in re.findall(regex_str, expression)]

    for key in new_keys:
        var_str = f"${type}{{{key}}}"
        var_info = variables.get(key)
        data_value = var_str
        if var_info is not None:
            if isinstance(var_info, dict):
                data_type = var_info.get("class")

                if data_type == "java.lang.Integer":
                    data_value = str(var_info.get("value", var_str))
                elif data_type == "java.sql.Timestamp":
                    date_time = var_info.get("value", var_str)
                    data_value = date_time
                    if isinstance(date_time, datetime.datetime):
                        data_value = (
                            "datetime.datetime("
                            + str(date_time.year)
                            + ","
                            + str(date_time.month)
                            + ","
                            + str(date_time.day)
                            + ","
                            + str(date_time.hour)
                            + ","
                            + str(date_time.minute)
                            + ","
                            + str(date_time.second)
                            + ","
                            + str(date_time.microsecond)
                            + ","
                            + str(date_time.tzinfo)
                            + ")"
                        )
                else:
                    is_blank_when_null = convert2boolean(attributes.get("isBlankWhenNull"))
                    value = var_info.get("value", var_str)
                    if is_blank_when_null and value is None:
                        data_value = ""
                    else:
                        data_value = '"' + str(var_info.get("value", var_str)) + '"'
            elif isinstance(var_info, str):
                data_value = str(var_info)

        expression = expression.replace(var_str, data_value)
    return expression


def replace_text(report, expression, attributes):
    """
    Replace Field and Variables with values ($F{} and $V{}) and evaluate expression.
    :report: dictionary holding report information
    :param report: dictionary holding report information
    :param expression: text element to evaluate
    :param attributes: attributes of text element to output
    :return: evaluated expression ready for output
    """
    expression = replace_fields(report, expression=expression, attributes=attributes)
    if isinstance(expression, (bytes, memoryview)):
        return expression
    expression = replace_variables(report.get("variables"), expression, attributes)
    expression = replace_parameters(report.get("parameters"), expression, attributes)

    try:
        return eval(expression)
    except SyntaxError:
        return expression
    except NameError:
        return expression


def strip_fname(name):
    """
    strip wrapper (e.g. F${}, V${}, P${}) from name to get ID.
    :param name: string to strip
    :return: element name without wrapper text
    """
    return name[3:-1]


def convert2boolean(text):
    """
    Convert a string to a boolean type.
    :param text: text to convert to boolean
    :return: 'True' if text is 'true' (case ignored). 'False' otherwise.
    """
    return text is not None and text.lower() == "true"



from PIL import Image
from reportlab.lib.utils import ImageReader
import io
import urllib.request

from pyametista.engine.bands.elements import process_report_element
from pyametista.engine.commonutilities import add_attr2attributes, replace_text
from pyametista.engine.components.line import process_box_element
from pyametista.engine.constants import protocols, workspace_prefix


import logging

logger = logging.getLogger(__name__)


def align_image(report, attributes, image):
    """
    Calculate x, y of an image based on horizontal alignment, vertical alignment, image size,
    and size specified in the jrxml.
    If the image size is larger, then image size is used.
    If the image size is smaller than the specified size, x and y coordinates are adjusted based on
    specified horizontal alignment and vertical alignment attribute values.
    :param report: dictionary holding report information
    :param attributes: attributes of an image to display
    :param image: image to display
    :return:
    """
    image_width, image_height = image.size
    x = attributes["x"]
    if attributes.get("hAlign") == "Right":
        if attributes["width"] > image_width:
            x += attributes["width"] - image_width
    elif attributes.get("hAlign") == "Center":
        if attributes["width"] > image_width:
            x += (attributes["width"] - image_width) / 2
    y = report["cur_y"] - attributes["y"] - image_height
    if attributes.get("vAlign") == "Bottom":
        if attributes["height"] > image_height:
            y -= attributes["height"] - image_height
    elif attributes.get("vAlign") == "Middle":
        if attributes["height"] > image_height:
            y -= (attributes["height"] - image_height) / 2
    return [x, y]


def draw_image(report, image, x, y, width, height):
    """
    Draw image on canvas.
    :param report: dictionary holding report information
    :param image: image to draw
    :param x: x coordinate to draw an image
    :param y: y coordinate to draw an image
    :param width: width of a drawn image
    :param height: height of a drawn image
    """
    im_data = io.BytesIO()
    image.save(im_data, format="png")
    im_data.seek(0)
    im_out = ImageReader(im_data)
    report["canvas"].drawImage(im_out, x, y, width=width, height=height, mask="auto")


def process_error(attributes, err):
    """
    Process image error based on 'On Error Type' value.
    :param attributes: attributes corresponding to the image
    :param err: except information
    """
    if attributes.get("onErrorType") == "Blank":  # do not print any image
        pass
    elif attributes.get("onErrorType") == "Icon":  # TODO support show icon on error
        logger.debug("show icon")
    else:
        logger.error(err)


def process_image_expression(report, element, attributes):
    """
    Process jrxml 'image_expression' element.
    :param report: dictionary holding report information
    :param element: jrxml image_expression element
    :param attributes:
    """
    image_expression = replace_text(report, element.get("value"), attributes)  # replace fields and variables
    if image_expression is not None:
        report["canvas"].saveState()

        try:
            if isinstance(image_expression, (bytes, memoryview)):
                stream = io.BytesIO(image_expression)
                image = Image.open(stream)
            else:
                image_expression = image_expression.strip('"')  # strip surrounding quotes
                if image_expression.startswith(protocols):
                    image_file = urllib.request.urlopen(image_expression)
                    image_expression = io.BytesIO(image_file.read())
                elif image_expression.startswith(workspace_prefix):
                    image_expression = image_expression[len(workspace_prefix) :]

                image = Image.open(image_expression)
            image_width, image_height = image.size

            # drawInlineImage(image, x, y, width=None, height=None, mask=None)
            if attributes.get("scaleImage") == "Clip":
                image_width, image_height = image.size
                width = min(image_width, attributes["width"])
                height = min(image_height, attributes["height"])
                im_crop = image.crop((0, 0, width, height))

                x, y = align_image(report, attributes, im_crop)
                # report['canvas'].drawInlineImage(im_crop, x, y, width=width, height=height)
                draw_image(report, im_crop, x, y, width, height)
            elif attributes.get("scaleImage") == "FillFrame":
                # report['canvas'].drawInlineImage(image, attributes['x'],
                #                                  report['cur_y'] - attributes['y'] - attributes['height'],
                #                                  width=attributes['width'], height=attributes['height'])
                draw_image(
                    report,
                    image,
                    attributes["x"],
                    report["cur_y"] - attributes["y"] - attributes["height"],
                    width=attributes["width"],
                    height=attributes["height"],
                )
            elif attributes.get("scaleImage") == "RealHeight":
                hsize = int(image_height * (attributes["width"] / float(image_width)))
                image = image.resize((attributes["width"], hsize), Image.ANTIALIAS)
                # report['canvas'].drawInlineImage(image, attributes['x'],
                #                                  report['cur_y'] - attributes['y'] - hsize)
                width, height = image.size
                draw_image(
                    report, image, attributes["x"], report["cur_y"] - attributes["y"] - hsize, width, height
                )

            elif attributes.get("scaleImage") == "RealSize":
                hsize = int(image_height * (attributes["width"] / float(image_width)))
                image = image.resize((attributes["width"], hsize), Image.ANTIALIAS)
                # report['canvas'].drawInlineImage(image, attributes['x'],
                #                                  report['cur_y'] - attributes['y'] - hsize)
                width, height = image.size
                draw_image(
                    report, image, attributes["x"], report["cur_y"] - attributes["y"] - hsize, width, height
                )
            else:  # reportElement.get('scaleImage') == 'RetainShape': RetainShape is the default behavior
                hsize = int(image_height * (attributes["width"] / float(image_width)))
                vsize = int(image_width * (attributes["height"] / float(image_height)))
                if hsize < vsize:
                    image_resize = image.resize((attributes["width"], hsize), Image.ANTIALIAS)
                elif hsize > vsize:
                    image_resize = image.resize((vsize, attributes["height"]), Image.ANTIALIAS)
                else:
                    image_resize = image.resize((hsize, vsize), Image.ANTIALIAS)  # Image.LANCZOS
                x, y = align_image(report, attributes, image_resize)
                # report['canvas'].drawInlineImage(image, x, y)
                width, height = image_resize.size
                draw_image(report, image_resize, x, y, width, height)

        except FileNotFoundError as err:
            # if attributes.get('onErrorType') == 'Blank':
            #     pass
            # elif attributes.get('onErrorType') == 'Icon':   # TODO support show icon on error
            #     logger.debug('show icon')
            # else:
            #     logger.error(err)
            process_error(attributes, err)
        except urllib.request.URLError as err:
            process_error(attributes, err)
        report["canvas"].restoreState()


"""
Possible elements under "image" jrxml element.
"""
image_dict = {
    "reportElement": process_report_element,
    "box": process_box_element,
    "graphicElement": None,
    "imageExpression": process_image_expression,
}


def process_image(report, element):
    """
    Process jrxml 'image' element.
    :param report: dictionary holding report information
    :param element: jrxml image element to process
    """
    image_element = element.get("child")
    if image_element is not None:
        report_element = process_report_element(
            report, image_element[0].get("reportElement")
        )  # get reportElement
        # get scaleImage attribute on image element
        add_attr2attributes(element, report_element)
        for tag in image_element[1:]:
            for key, value in tag.items():
                if image_dict[key] is not None:
                    image_dict[key](report, value, report_element)



import csv
import logging

logger = logging.getLogger(__name__)


class CSVAdapter:
    """
    CSV datasource adapter.
    """
    def __init__(self, filename):
        try:
            self.csv_file = open(filename, 'r')
            self.csv_reader = csv.reader(self.csv_file)
            # self.csv_reader = csv.DictReader(self.csv_file)
        except IOError as err:
            if 'csv_file' in locals():
                self.csv_file.close()
            logger.error('Unable to open specified csv file', exc_info=True)

    def get_cursor(self):
        """
        Return next row in query execution.
        execute_query() should be executed before calling this method.
        :return: next query result row.
        """
        return next(self.csv_reader)
        # return None

    def execute_query(self, query_string):
        """
        Query a database table.
        :param query_string: query string to execute.
        :return: list of column names.
        """
        if self.csv_reader is None:
            return None
        else:
            return next(self.csv_reader)

    def fetch_row(self):
        if self.csv_reader is None:
            return None
        else:
            try:
                return next(self.csv_reader)
            except StopIteration as err:
                return None
            except csv.Error:
                return None

    def close_cursor(self):
        """
        Close query execution cursor and connection.
        """
        if 'csv_file' in locals():
            self.csv_file.close()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)

    # input_filename = '../../tests/data/address.csv'
    data_config = {'adapter': 'csv', 'filename': '../../tests/data/address.csv'}

    adapter = CSVAdapter(data_config.get('filename'))
    header = adapter.execute_query('')
    logger.info(header)
    adapter.close_cursor()




from pyametista.engine.bands.elements import (process_graphic_element,
                                              process_report_element)
from pyametista.engine.commonutilities import add_attr2attributes
from reportlab.lib import colors

"""
List of possible line styles.
"""
line_style_dict = {"Solid": (1), "Dashed": (6, 3), "Dotted": (1, 2), "Double": (1)}


def set_pen_attr(element, attributes):
    """
    Set pen attributes.
    :param element:
    :param attributes:
    :return:
    """
    pen_attr = element.get("attr")
    if pen_attr is None:
        return {}
    else:
        return {
            "lineStyle": pen_attr.get("lineStyle"),
            "lineWidth": float(pen_attr.get("lineWidth", "0")),
            "lineColor": pen_attr.get("lineColor"),
        }


def draw_line(report, x1, y1, x2, y2, line_style, line_width, line_color):
    """
    Draw a line on report_info.
    :param report: dictionary holding report information
    :param x1: start x coordinate
    :param y1: start y coordinate
    :param x2: end x coordinate
    :param y2: end y coordinate
    :param line_style: line style
    :param line_width: line width
    :param line_color: line color in #FFFFFF format
    """
    report["canvas"].saveState()

    report["canvas"].setDash(line_style_dict.get(line_style, (1)))
    if line_color is not None:
        report["canvas"].setStrokeColor(colors.HexColor(line_color))

    if line_style != "Double":  # double line consists of wide solid line with thin white line in between
        report["canvas"].setLineWidth(line_width)
        report["canvas"].line(x1, y1, x2, y2)
    else:
        report["canvas"].setLineWidth(line_width * 1.5)
        report["canvas"].line(x1, y1, x2, y2)
        report["canvas"].setLineWidth(line_width * 0.7)
        report["canvas"].setStrokeColor(colors.HexColor("#FFFFFF"))
        report["canvas"].line(x1, y1, x2, y2)

    report["canvas"].restoreState()


def draw_diagonal_line(report, attributes, line_style, line_width, line_color):
    """
    Process to jrxml diagonal line.
    :param report: dictionary holding report information
    :param attributes: line attributes. Attributes from "reportElement" element.
    :param line_style: line style
    :param line_width: line width
    :param line_color: line color
    """
    if attributes["width"] == 1:
        width = 0
    else:
        width = attributes["width"]

    if attributes["height"] == 1:
        height = 0
    else:
        height = attributes["height"]

    if attributes.get("direction") is None:  # top down
        draw_line(
            report,
            attributes["x"],
            report["cur_y"] - attributes["y"],
            attributes["x"] + width,
            report["cur_y"] - attributes["y"] - height,
            line_style,
            line_width,
            line_color,
        )
    else:  # bottom up
        draw_line(
            report,
            attributes["x"],
            report["cur_y"] - attributes["y"] - height,
            attributes["x"] + width,
            report["cur_y"] - attributes["y"],
            line_style,
            line_width,
            line_color,
        )


def process_box_element(report, element, attributes):
    """
    Draw borders around a component.
    :param report: dictionary holding report information
    :param element: jrxml box element
    :param attributes:
    """
    # global report_info

    report["canvas"].saveState()

    borders = element.get("child")
    if borders is not None:
        cell_border = {}
        # find each border line attributes
        for pen_elements in borders:
            for pen_key, pen_value in pen_elements.items():
                if pen_key == "pen":  # draw all borders
                    pen = set_pen_attr(pen_value, attributes)
                    if pen["lineWidth"] > 0:  # all borders
                        cell_border["top"] = pen
                        cell_border["left"] = pen
                        cell_border["bottom"] = pen
                        cell_border["right"] = pen
                elif pen_key == "topPen":
                    pen = set_pen_attr(pen_value, attributes)
                    if pen["lineWidth"] > 0:
                        cell_border["top"] = pen
                elif pen_key == "leftPen":
                    pen = set_pen_attr(pen_value, attributes)
                    if pen["lineWidth"] > 0:
                        cell_border["left"] = pen
                elif pen_key == "bottomPen":
                    pen = set_pen_attr(pen_value, attributes)
                    if pen["lineWidth"] > 0:
                        cell_border["bottom"] = pen
                elif pen_key == "rightPen":
                    pen = set_pen_attr(pen_value, attributes)
                    if pen["lineWidth"] > 0:
                        cell_border["right"] = pen
        # draw each border lines
        # draw top
        if cell_border.get("top") is not None:
            draw_line(
                report,
                attributes["x"],
                report["cur_y"] - attributes["y"],
                attributes["x"] + attributes["width"],
                report["cur_y"] - attributes["y"],
                cell_border["top"]["lineStyle"],
                cell_border["top"]["lineWidth"],
                cell_border["top"]["lineColor"],
            )
        # draw left
        if cell_border.get("left") is not None:
            draw_line(
                report,
                attributes["x"],
                report["cur_y"] - attributes["y"],
                attributes["x"],
                report["cur_y"] - attributes["y"] - attributes["height"],
                cell_border["left"]["lineStyle"],
                cell_border["left"]["lineWidth"],
                cell_border["left"]["lineColor"],
            )
        # draw bottom
        if cell_border.get("bottom") is not None:
            draw_line(
                report,
                attributes["x"],
                report["cur_y"] - attributes["y"] - attributes["height"],
                attributes["x"] + attributes["width"],
                report["cur_y"] - attributes["y"] - attributes["height"],
                cell_border["bottom"]["lineStyle"],
                cell_border["bottom"]["lineWidth"],
                cell_border["bottom"]["lineColor"],
            )
        # draw right
        if cell_border.get("right") is not None:
            draw_line(
                report,
                attributes["x"] + attributes["width"],
                report["cur_y"] - attributes["y"],
                attributes["x"] + attributes["width"],
                report["cur_y"] - attributes["y"] - attributes["height"],
                cell_border["right"]["lineStyle"],
                cell_border["right"]["lineWidth"],
                cell_border["right"]["lineColor"],
            )
    report["canvas"].restoreState()


def process_line(report, element):
    """
    Process jrxml line element.
    :param report: dictionary holding report information
    :param element: jrxml line element
    """
    line_element = element.get("child")
    if line_element is not None:
        report_element = process_report_element(
            report, line_element[0].get("reportElement")
        )  # get reportElement
        # get attributes on line element (e.g. "direction") to determine which direction to draw a line
        add_attr2attributes(element, report_element)

        # set attributes from following graphicElement
        # reportElement = process_graphic_element(report, line_element, report_element)
        process_graphic_element(report, line_element, report_element)

        # set line attributes
        line_style = report_element.get("lineStyle")
        line_width = float(report_element.get("lineWidth", "1.0"))
        line_color = report_element.get("lineColor")

        draw_diagonal_line(report, report_element, line_style, line_width, line_color)



import logging

import psycopg2

from pyametista.engine.bands.elements import add_attr2attributes
from pyametista.engine.components.components import process_componentElement
from pyametista.engine.components.image import process_image
from pyametista.engine.components.line import process_line
from pyametista.engine.components.shapes import process_circle, process_ellipse, process_rectangle
from pyametista.engine.components.text import process_static_text, process_text_field
from pyametista.exports.pdf import create_canvas, write_to_file

logger = logging.getLogger(__name__)


def create_new_page(report):
    """
    End the current page and create a new page.
    :param report: dictionary holding report information
    """
    report["printingFooter"] = True
    column_footer_band(report)
    page_footer_band(report)
    report["printingFooter"] = False
    report["canvas"].showPage()
    report["cur_y"] = report["pageHeight"] - report["topMargin"]
    page_header_band(report=report, force=True)
    column_header_band(report=report, force=True)
    report["variables"]["PAGE_NUMBER"]["value"] += 1
    report["variables"]["PAGE_COUNT"]["value"] = 1
    report["variables"]["COLUMN_COUNT"]["value"] = 1


def process_property_band(report, element):
    """
    Read jrxml properties elements and set to 'properties' dictionary.
    :param report: dictionary holding report information
    :param properties:
    :param element: jrxml properties element
    """
    # return add_attr2attributes(element, properties)
    add_attr2attributes(element, report.get("properties"))


def process_query_string_band(report, element):
    """
    Read from jrxml queryString element and query a datasource.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    """
    sql_stmt = element.get("value")
    if sql_stmt is not None:
        sql_stmt = sql_stmt.strip()
        # TODO instead of getting field names from query, get from 'field' element list
        report["main_datasource"].execute_query(sql_stmt)
        # instead of getting field names from datasourvce, get field names field elements
        # report['field_names'] = report['main_datasource'].execute_query(sql_stmt)


def process_field_band(report, element):
    """
    Process jrxml field element.
    Add field names and it's class (datatype) info to 'field_dict' in global 'report_info' variable.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    """
    field_attr = element.get("attr")
    name = field_attr.get("name")
    datatype = field_attr.get("class")

    if report.get("field_dict") is None:
        report["field_dict"] = dict()
    report["field_dict"][name] = datatype

    if report.get("field_names") is None:
        report["field_names"] = [name]
    else:
        report["field_names"].append(name)


def process_variable_band(report, element):
    """
    Process jrxml variable element.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    """
    pass  # currently not supported


def process_background_band(report, element):
    """
    Process jrxml background element.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    """
    pass  # currently not supported


def process_group_expression(report, element):
    """
    Process groupExpression element.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    :return:
    """
    if element is not None:
        # group_name = report['group_cur'][-1]
        report["group_names"].update({report["group_cur"][-1]: element.get("value")})


def process_group_header(report, element):
    """
    Process group header element.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    """
    if element is not None:
        if (
            report.get("main_datasource") is None
        ):  # when there is no datasource, just process static page
            # row_data = None
            check_for_new_page(report)

            process_band_element(report, element)
        else:
            check_for_new_page(report)
            report_elements_list.remove("detail")  # call details from below while loop
            while True:
                row = report["main_datasource"].fetch_row()
                if (
                    row is None
                ):  # there is no data in datasource then just print content in jrxml layout
                    if not report["output_to_canvas"]:
                        check_for_new_page(report)
                        process_band_element(report, element)
                        report["output_to_canvas"] = True
                    break
                elif report["group_names"].get(report["group_cur"][-1]) is None:
                    check_for_new_page(report)
                    process_datasource_row(row, report)
                    process_band_element(report, element)
                    report["output_to_canvas"] = True

                    detail_band = report.get("detail")
                    if detail_band is not None and len(detail_band) > 0:
                        detail = detail_band[0].get("detail")
                        detail_band_child = detail.get("child")
                        if detail_band_child is not None:
                            detail_element = detail_band_child[0]
                            # loop through detail band
                            while row is not None:
                                process_band(report, detail_element.get("band"))
                                row = fetch_from_datasource(report)
                    break
                else:  # if there is some row in datasource
                    check_for_new_page(report)
                    process_datasource_row(row, report)
                    # check if there is a prev value or if there is current group
                    if not report["prev_value"] or report["group_cur"] is None:
                        process_band_element(report, element)
                        report["prev_value"] = report["row_data"]
                        report["output_to_canvas"] = True
                    else:
                        cur_group_value = report["row_data"].get(report["group_cur"][-1])
                        if report["prev_value"].get(report["group_cur"][-1]) != cur_group_value:
                            process_band_element(report, element)
                            report["prev_value"] = report["row_data"]
                            report["output_to_canvas"] = True


def process_group_footer(report, element):
    """
    Process groupFooter element.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    """
    if element is not None:
        if report.get("canvas") is None:
            create_canvas(report)
        process_band_element(report, element)


"""supported jrxml elements under group element"""
group_element_dict = {
    "groupExpression": process_group_expression,
    "groupHeader": process_group_header,
    "groupFooter": process_group_footer,
}


def process_group_band(report, element):
    """
    Process group element.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    """
    group_attr = element.get("attr")
    report["group_names"].update({group_attr.get("name"): None})
    report["group_cur"].append(group_attr.get("name"))
    # report['group_cur'] = group_attr.get('name')

    if element.get("child") is not None:
        for tag in element.get("child"):
            for key, value in tag.items():
                if group_element_dict.get(key) is not None:
                    group_element_dict[key](report, value)
    report["group_cur"].pop()


def process_title_band(report, element=None):
    """
    Process title element.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    """
    if not report["output_to_canvas"]:
        if report.get("canvas") is None:
            create_canvas(report)
        process_band_element(report, element)


def page_header_band(report, element=None, force=False):
    """
    Process pageHeader element.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    :param force: True if unconditionally output, False if only output if nothing is yet output to canvas
    """
    if not report["output_to_canvas"] or force:
        if element is None:
            page_header = report.get("pageHeader")
            if page_header is not None:
                element = report["pageHeader"][0]["pageHeader"]

        if report.get("canvas") is None:
            create_canvas(report)
        process_band_element(report, element)


def column_header_band(report, element=None, force=False):
    """
    Process columnHeader element.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    :param force: True if unconditionally output, False if only output if nothing is yet output to canvas
    """
    if not report["output_to_canvas"] or force:
        if element is None:
            column_header = report.get("columnHeader")
            if column_header is not None:
                element = report["columnHeader"][0]["columnHeader"]
        if report.get("canvas") is None:
            create_canvas(report)
        process_band_element(report, element)


def get_band_height(element):
    """
    Return height the the specified element.
    :param element: current jrxml element being processes.
    :return:
    """
    return float(element["child"][0]["band"]["attr"]["height"])


def calc_column_footer_band_height(report):
    """
    Calculate column footer band height.
    :param report: dictionary holding report information
    :return: height of column footer band.
    """
    if report.get("columnFooter") is None:
        return 0
    else:
        return get_band_height(report["columnFooter"][0]["columnFooter"])


def column_footer_band(report, element=None):
    """
    Process columnFooter element.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    :return:
    """
    if element is None:
        column_footer = report.get("columnFooter")
        if column_footer is not None:
            element = report["columnFooter"][0]["columnFooter"]
    if report.get("canvas") is None:
        create_canvas(report)
    report["printingFooter"] = True
    process_band_element(report, element)


def calc_page_footer_band_height(report):
    """
    Calculate pageFooter band height.
    :param report: dictionary holding report information
    :return: page footer band height
    """
    if report.get("pageFooter") is None:
        return 0
    else:
        return get_band_height(report["pageFooter"][0]["pageFooter"])


def page_footer_band(report, element=None):
    """
    Process jrxml 'pageFooter' element.
    :param report: dictionary holding report information
    :param element: parent element of jrxml 'pageFooter' element
    """
    if element is None:
        page_footer = report.get("pageFooter")
        if page_footer is not None:
            element = report["pageFooter"][0]["pageFooter"]
    if report.get("canvas") is None:
        create_canvas(report)
    report["printingFooter"] = True
    process_band_element(report, element)


def summary_band(report, element):
    """
    Process jrxml 'summmary' element.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    """
    if report.get("canvas") is None:
        create_canvas(report)
    report["printingFooter"] = True
    process_band_element(report, element)


def process_band_element(report, element):
    """
    Process jrxml band element.
    :param report: dictionary holding report information
    :param element: jrxml band element
    """
    if element is not None:
        band_element = element.get("child")
        if band_element is not None:
            title_element = band_element[0]
            process_band(report, title_element.get("band"))


"""Elements in jrxml 'band' element"""
band_elements_dict = {
    "printWhenExpression": None,
    "break": None,
    "line": process_line,
    "rectangle": process_rectangle,
    "ellipse": process_ellipse,
    "circle": process_circle,
    "image": process_image,
    "staticText": process_static_text,
    "textField": process_text_field,
    "frame": None,
    "componentElement": process_componentElement,
}


def process_band(report, element):
    """
    Process jrxml band element. Page breaks are inserted when there is not enough space to write a band.
    :param report: dictionary holding report information
    :param element: current jrxml element being processes.
    """
    band_settings = element.get("attr")
    band_height = int(band_settings.get("height", "34"))

    # band_split_type = band_settings.get('splitType', 'Stretch')
    if (
        report["cur_y"]
        < (
            report["bottomMargin"]
            + report["col_footer_height"]
            + report["page_footer_height"]
            + band_height
        )
        and not report["printingFooter"]
    ):
        create_new_page(report)

    if element.get("child") is not None:
        for tag in element.get("child"):
            for key, value in tag.items():
                if band_elements_dict.get(key) is not None:
                    # band_elements_dict[key](report=report, element=value, row_data=row_data)
                    band_elements_dict[key](report=report, element=value)

    report["cur_y"] = report["cur_y"] - band_height


# def fetch_row_data(report, element, row):
#     """
#     Loop through datasource and output detail band rows
#     :param element: 'detail' jrxml element
#     """
#
#             while True:
#                 if report['main_datasource'] is not None:
#                     row = report['main_datasource'].fetch_row()
#
#                     report['variables']['REPORT_COUNT']['value'] += 1
#                     report['variables']['PAGE_COUNT']['value'] += 1
#
#                     if row is None:
#                         break
#                     report['variables']['COLUMN_COUNT']['value'] += 1
#
#                     if type(row) == tuple or type(row) == list:  # regular result set
#                         # TODO instead of creating a dictionary, directly use list instead.
#                         #  Replace using Field elements entries
#                         row_data = dict((field, value) for field, value in zip(report['field_names'], row))
#                     elif type(row) == psycopg2.extras.DictRow:  # result in dictionary format
#                         row_data = row
#                     else:
#                         logger.error("invalid datasource format.")
#                         row_data = None
#
#                     process_band(report, title_element.get('band'), row_data)


def check_for_new_page(report):
    if report.get("canvas") is None:
        create_canvas(report)
        if report.get("title") is not None:
            report_elements_dict["title"](report, report["title"][0].get("title"))
        if report.get("pageHeader") is not None:
            report_elements_dict["pageHeader"](report, report["pageHeader"][0].get("pageHeader"))
        if report.get("columnHeader") is not None:
            report_elements_dict["columnHeader"](
                report, report["columnHeader"][0].get("columnHeader")
            )


def process_datasource_row(row, report):
    report["row_data"] = row
    if type(row) == tuple or type(row) == list:
        report["row_data"] = dict(
            (field, value) for field, value in zip(report["field_names"], row)
        )


def fetch_from_datasource(report):
    row = report["main_datasource"].fetch_row()

    report["variables"]["REPORT_COUNT"]["value"] += 1
    report["variables"]["PAGE_COUNT"]["value"] += 1

    # if row is None:
    #     break
    if row is not None:
        report["variables"]["COLUMN_COUNT"]["value"] += 1
        process_datasource_row(row, report)
    return row


def process_detail_band_element(report, element):
    """
    Loop through datasource and output detail band rows
    :param report: dictionary holding report information
    :param element: 'detail' jrxml element
    """

    band_element = element.get("child")
    if band_element is not None:  # only process if there is a jrxml 'band' element
        detail_element = band_element[0]

        # check_for_new_page(report)

        if (
            report.get("main_datasource") is None
        ):  # when there is no datasource, just process static page
            row_data = None
            check_for_new_page(report)

            # process_band(report, title_element.get('band'), row_data)

            process_band(report, detail_element.get("band"))
        else:
            # fetch row from datasource and process each row
            check_for_new_page(report)
            row = fetch_from_datasource(report)
            while row is not None:
                # if fetch_from_datasource(report) is None:
                #     break;
                # if report['main_datasource'] is not None:
                # row = report['main_datasource'].fetch_row()
                #
                # report['variables']['REPORT_COUNT']['value'] += 1
                # report['variables']['PAGE_COUNT']['value'] += 1
                #
                # if row is None:
                #     break
                # report['variables']['COLUMN_COUNT']['value'] += 1
                #
                # process_datasource_row(row, report)
                process_band(report, detail_element.get("band"))
                row = fetch_from_datasource(report)
                # TODO if page per record, then set report['canvas'] = None

                # if type(row) == tuple or type(row) == list:  # regular result set
                #     # TODO instead of creating a dictionary, directly use list instead.
                #     #  Replace using Field elements entries
                #     report['row_data'] = dict((field, value) for field, value in zip(report['field_names'], row))
                # elif type(row) == psycopg2.extras.DictRow:  # result in dictionary format
                #     report['row_data'] = row
                # else:
                #     logging.error("invalid datasource format.")
                #     report['row_data'] = None
                # process_band(report, detail_element.get('band'))
                # # TODO if page per record, then set report['canvas'] = None
    report[
        "output_to_canvas"
    ] = True  # TODO this needs to be moved to code actually writing to canvas


def process_last_page_footer(report, element):
    if report.get("canvas") is None:
        create_canvas(report)
    report["printingFooter"] = True
    process_band_element(report, element)


def process_no_data(report, element):
    if report.get("canvas") is None:
        create_canvas(report)
    report["printingFooter"] = True
    process_band_element(report, element)


"""
list of report_info elements in a report_info.
"""
report_elements_dict = {
    "property": process_property_band,
    "import": None,
    "template": None,
    "reportFont": None,
    "style": None,
    "subDataset": None,
    "parameter": None,
    "queryString": process_query_string_band,
    "field": process_field_band,
    "sortField": None,
    "variable": process_variable_band,
    "filterExpression": None,
    "group": process_group_band,
    "background": process_background_band,
    "title": process_title_band,
    "pageHeader": page_header_band,
    "columnHeader": column_header_band,
    "detail": process_detail_band_element,
    "columnFooter": column_footer_band,
    "pageFooter": page_footer_band,
    "lastPageFooter": process_last_page_footer,
    "summary": summary_band,
    "noData": process_no_data,
}

"""
Define order to process tags. Tags in jrxml are not in order that they are displayed in a report_info.
For example, 'Summary' band appears before 'Column Footer' band.
"""
report_elements_list = [
    "property",
    "import",
    "template",
    "reportFont",
    "style",
    "subDataset",
    "parameter",
    "queryString",
    "field",
    "sortField",
    "variable",
    "filterExpression",
    "background",
    # "title",          # process title, pageHeader, columnHeader after later to make it possible to have
    # "pageHeader",     # filename based on datasource
    # "columnHeader",
    "group",  # TODO currently not supported. Need to read datasource in group
    "detail",
    "summary",
    "columnFooter",
    "title",
    "pageHeader",
    "columnHeader",
    # "group",
    # "pageFooter",     # pageFooter, lastPageFooter, and noDate are processed separately because they are either one.
    # "lastPageFooter",
    # 'noData'
]


def process_no_pages(report_info):
    """
    Do nothing here. Saving to file is skipped in generate_pdf() method.
    """
    pass


def process_blank_page(report):
    """
    Re-initialize report_info and output blank page.
    :param report: dictionary holding report information
    """
    # create_canvas(report, report['output_filename'])    # clear report_info
    create_canvas(report)  # clear report_info


def process_all_sections_no_detail(report):
    """
    Output Title, Page Header, Column Header, Summary, Column Footer, Page Footer.
    Since normal processing will output all these, just pass through.
    :param report: dictionary holding report information
    """
    pass


def process_no_data_section(report):
    """
    Re-initialize report_info and output 'No Data' band only.
    :param report: dictionary holding report information
    """
    # create_canvas(report, report['output_filename'])    # clear report_info
    create_canvas(report)  # clear report_info

    report["cur_y"] = report["pageHeight"] - report["topMargin"]
    if report.get("noData") is not None:  # print noData band if defined in layout
        report_elements_dict["noData"](report, report["noData"][0].get("noData"))


def process_null_data(report):
    pass


"""Dictionary of method to execute when there is no data in datasource."""
when_no_data_type_dict = {
    "Null": process_null_data,
    "NoPages": process_no_pages,
    "NoDataSection": process_no_data_section,
    "BlankPage": process_blank_page,
    "AllSectionsNoDetail": process_all_sections_no_detail,
    "NoDataSection": process_no_data_section,
}


def process_bands(report, root):
    """
    Process main report_info elements (bands).
    Order and which element to processed is defined in 'report_elements_list'.
    Methods to execute for each element is defined in 'report_elements_dict'.
    :param report: dictionary holding report information
    :param root: jrxml 'jasperReport' element
    """
    for tag in report_elements_list:  # loop through jrxml elements to be processed
        element = root.get(tag)
        if element is not None:
            for element_tag in element:  # loop through jrxml elements
                for key, value in element_tag.items():
                    if (
                        report_elements_dict[key] is not None
                    ):  # process only if currently supported by pyAmetista
                        report_elements_dict[key](report=report, element=value)
    if report.get("lastPageFooter") is not None:  # print lastPageFooter if defined in layout
        report_elements_dict["lastPageFooter"](
            report, report["lastPageFooter"][0].get("lastPageFooter")
        )
    elif report.get("pageFooter") is not None:  # else if pageFooter is defined, print it out
        report_elements_dict["pageFooter"](report, report["pageFooter"][0].get("pageFooter"))

    # close database cursor/connection
    if report.get("main_datasource") is not None:
        report["main_datasource"].close_cursor()

    # if no datasource row was processed and when 'When No Data Type' is <NULL> or 'No Pages',
    # then do not create a report_info. Otherwise, create a report_info file.
    if report["main_datasource"] is not None and report["variables"]["COLUMN_COUNT"]["value"] < 1:
        when_no_data_type_dict[report.get("whenNoDataType")](report)
        if report.get("whenNoDataType") != "NoPages":
            write_to_file(report)
    else:
        write_to_file(report)  # write report_info to specified file.



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyametista',
 'pyametista.adapters',
 'pyametista.engine',
 'pyametista.engine.bands',
 'pyametista.engine.components',
 'pyametista.exports']

package_data = \
{'': ['*'],
 'pyametista': ['fonts/LICENSE.txt',
                'fonts/LICENSE.txt',
                'fonts/LICENSE.txt',
                'fonts/LICENSE.txt',
                'fonts/LICENSE.txt',
                'fonts/LICENSE.txt',
                'fonts/LICENSE.txt',
                'fonts/LICENSE.txt',
                'fonts/LICENSE.txt',
                'fonts/LICENSE.txt',
                'fonts/LICENSE.txt',
                'fonts/OpenSans-Bold.ttf',
                'fonts/OpenSans-Bold.ttf',
                'fonts/OpenSans-Bold.ttf',
                'fonts/OpenSans-Bold.ttf',
                'fonts/OpenSans-Bold.ttf',
                'fonts/OpenSans-Bold.ttf',
                'fonts/OpenSans-Bold.ttf',
                'fonts/OpenSans-Bold.ttf',
                'fonts/OpenSans-Bold.ttf',
                'fonts/OpenSans-Bold.ttf',
                'fonts/OpenSans-Bold.ttf',
                'fonts/OpenSans-BoldItalic.ttf',
                'fonts/OpenSans-BoldItalic.ttf',
                'fonts/OpenSans-BoldItalic.ttf',
                'fonts/OpenSans-BoldItalic.ttf',
                'fonts/OpenSans-BoldItalic.ttf',
                'fonts/OpenSans-BoldItalic.ttf',
                'fonts/OpenSans-BoldItalic.ttf',
                'fonts/OpenSans-BoldItalic.ttf',
                'fonts/OpenSans-BoldItalic.ttf',
                'fonts/OpenSans-BoldItalic.ttf',
                'fonts/OpenSans-BoldItalic.ttf',
                'fonts/OpenSans-Italic.ttf',
                'fonts/OpenSans-Italic.ttf',
                'fonts/OpenSans-Italic.ttf',
                'fonts/OpenSans-Italic.ttf',
                'fonts/OpenSans-Italic.ttf',
                'fonts/OpenSans-Italic.ttf',
                'fonts/OpenSans-Italic.ttf',
                'fonts/OpenSans-Italic.ttf',
                'fonts/OpenSans-Italic.ttf',
                'fonts/OpenSans-Italic.ttf',
                'fonts/OpenSans-Italic.ttf',
                'fonts/OpenSans-Regular.ttf',
                'fonts/OpenSans-Regular.ttf',
                'fonts/OpenSans-Regular.ttf',
                'fonts/OpenSans-Regular.ttf',
                'fonts/OpenSans-Regular.ttf',
                'fonts/OpenSans-Regular.ttf',
                'fonts/OpenSans-Regular.ttf',
                'fonts/OpenSans-Regular.ttf',
                'fonts/OpenSans-Regular.ttf',
                'fonts/OpenSans-Regular.ttf',
                'fonts/OpenSans-Regular.ttf',
                'fonts/Roboto-Bold.ttf',
                'fonts/Roboto-Bold.ttf',
                'fonts/Roboto-Bold.ttf',
                'fonts/Roboto-Bold.ttf',
                'fonts/Roboto-Bold.ttf',
                'fonts/Roboto-Bold.ttf',
                'fonts/Roboto-Bold.ttf',
                'fonts/Roboto-Bold.ttf',
                'fonts/Roboto-Bold.ttf',
                'fonts/Roboto-Bold.ttf',
                'fonts/Roboto-Bold.ttf',
                'fonts/Roboto-BoldItalic.ttf',
                'fonts/Roboto-BoldItalic.ttf',
                'fonts/Roboto-BoldItalic.ttf',
                'fonts/Roboto-BoldItalic.ttf',
                'fonts/Roboto-BoldItalic.ttf',
                'fonts/Roboto-BoldItalic.ttf',
                'fonts/Roboto-BoldItalic.ttf',
                'fonts/Roboto-BoldItalic.ttf',
                'fonts/Roboto-BoldItalic.ttf',
                'fonts/Roboto-BoldItalic.ttf',
                'fonts/Roboto-BoldItalic.ttf',
                'fonts/Roboto-Italic.ttf',
                'fonts/Roboto-Italic.ttf',
                'fonts/Roboto-Italic.ttf',
                'fonts/Roboto-Italic.ttf',
                'fonts/Roboto-Italic.ttf',
                'fonts/Roboto-Italic.ttf',
                'fonts/Roboto-Italic.ttf',
                'fonts/Roboto-Italic.ttf',
                'fonts/Roboto-Italic.ttf',
                'fonts/Roboto-Italic.ttf',
                'fonts/Roboto-Italic.ttf',
                'fonts/Roboto-Regular.ttf',
                'fonts/Roboto-Regular.ttf',
                'fonts/Roboto-Regular.ttf',
                'fonts/Roboto-Regular.ttf',
                'fonts/Roboto-Regular.ttf',
                'fonts/Roboto-Regular.ttf',
                'fonts/Roboto-Regular.ttf',
                'fonts/Roboto-Regular.ttf',
                'fonts/Roboto-Regular.ttf',
                'fonts/Roboto-Regular.ttf',
                'fonts/Roboto-Regular.ttf',
                'fonts/RobotoSlab-Bold.ttf',
                'fonts/RobotoSlab-Bold.ttf',
                'fonts/RobotoSlab-Bold.ttf',
                'fonts/RobotoSlab-Bold.ttf',
                'fonts/RobotoSlab-Bold.ttf',
                'fonts/RobotoSlab-Bold.ttf',
                'fonts/RobotoSlab-Bold.ttf',
                'fonts/RobotoSlab-Bold.ttf',
                'fonts/RobotoSlab-Bold.ttf',
                'fonts/RobotoSlab-Bold.ttf',
                'fonts/RobotoSlab-Bold.ttf',
                'fonts/RobotoSlab-Regular.ttf',
                'fonts/RobotoSlab-Regular.ttf',
                'fonts/RobotoSlab-Regular.ttf',
                'fonts/RobotoSlab-Regular.ttf',
                'fonts/RobotoSlab-Regular.ttf',
                'fonts/RobotoSlab-Regular.ttf',
                'fonts/RobotoSlab-Regular.ttf',
                'fonts/RobotoSlab-Regular.ttf',
                'fonts/RobotoSlab-Regular.ttf',
                'fonts/RobotoSlab-Regular.ttf',
                'fonts/RobotoSlab-Regular.ttf']}

install_requires = \
['mysql-connector-python>=8.0.28,<9.0.0',
 'psycopg2-binary>=2.9.3,<3.0.0',
 'reportlab>=3.6.8,<4.0.0']

setup_kwargs = {
    'name': 'pyametista',
    'version': '0.1.3',
    'description': 'Módulo Python para geração de PDF utilizando arquivos .jrxml (JasperReports)',
    'long_description': None,
    'author': 'Bruno Souza',
    'author_email': 'bruno@komu.com.br',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from reportlab.lib.pagesizes import A4
from reportlab.pdfgen import canvas

# TODO this should be made into a class with all write operations as a class method

def create_canvas(report):
    """
    Create a new reportLab canvas to generate pdf.
    :param report: dictionary holding report information

    """
    default_page_width, default_page_height = A4    # set default page size to A4
    global page_width, page_height, cur_y

    page_width= report.get('pageWidth', default_page_width)
    page_height = report.get('pageHeight', default_page_height)
    cur_y = page_height

    report['canvas'] = canvas.Canvas(report['output_filename'], pagesize=(page_width, page_height))
    report['canvas'].setAuthor('pyAmetista')
    report['canvas'].setTitle(report.get('name', 'sample report_info'))
    report['canvas'].setSubject(report.get('name', 'sample report_info'))


def write_to_file(report):
    """
    Start a new page and save current report_info to a file.
    :param report: dictionary holding report information
    """
    # save page to file
    if report.get('canvas') is not None:
        report['canvas'].showPage()  # start a new page
        report['canvas'].save()      # save to a file



from reportlab.lib import colors

from pyametista.engine.bands.elements import process_graphic_element, process_report_element
from pyametista.engine.commonutilities import add_attr2attributes
from pyametista.engine.components.line import line_style_dict


def draw_rectangle(report, attributes, line_style, line_width, fore_color, background_color, fill):
    """
    Draw a rectangle on report_info.
    :param attributes:
    :param line_style:
    :param line_width:
    :param fore_color:
    :param background_color:
    :param fill:
    """
    # global report_info

    report["canvas"].saveState()

    report["canvas"].setDash(line_style_dict.get(line_style, (1)))
    report["canvas"].setLineWidth(line_width)
    if fore_color is not None:
        report["canvas"].setStrokeColor(colors.HexColor(fore_color))
    if background_color is not None:
        report["canvas"].setFillColor(colors.HexColor(background_color))
    if attributes.get("radius") is None:
        report["canvas"].rect(
            attributes["x"],
            report["cur_y"] - attributes["y"] - attributes["height"],
            attributes["width"],
            attributes["height"],
            fill=fill,
        )
    else:
        report["canvas"].roundRect(
            attributes["x"],
            report["cur_y"] - attributes["y"] - attributes["height"],
            attributes["width"],
            attributes["height"],
            radius=attributes.get("radius"),
            fill=fill,
        )

    report["canvas"].restoreState()


def process_rectangle(report, element):
    """
    Process jrxml "rectangle" element.
    :param report:
    :param element:
    :return:
    """
    # global report_info

    rectangle_element = element.get("child")
    if rectangle_element is not None:
        attributes = process_report_element(
            report, rectangle_element[0].get("reportElement")
        )  # get reportElement
        # get radius attribute on rectangle element
        add_attr2attributes(element, attributes)

        # set attributes from following graphicElement
        attributes = process_graphic_element(report, rectangle_element, attributes)

        # set line attributes
        line_style = attributes.get("lineStyle")
        line_width = float(attributes.get("lineWidth", "1.0"))
        fore_color = attributes.get("forecolor")
        background_color = attributes.get("backcolor")

        fill = 0
        if background_color is not None:
            fill = 1

        # draw rectangle
        if (
            line_style == "Double"
        ):  # double line consists of wide solid line with thin white line in between
            draw_rectangle(
                report,
                attributes,
                line_style_dict["Solid"],
                line_width * 1.5,
                fore_color,
                background_color,
                fill,
            )
            draw_rectangle(
                report,
                attributes,
                line_style_dict["Solid"],
                line_width * 0.3,
                "#FFFFFF",
                background_color,
                fill,
            )
        else:
            draw_rectangle(
                report, attributes, line_style, line_width, fore_color, background_color, fill
            )


def draw_ellipse(report, attributes, line_style, line_width, fore_color, background_color, fill):
    """
    Draw an ellipse on report_info.
    :param report:
    :param attributes:
    :param line_style:
    :param line_width:
    :param fore_color:
    :param background_color:
    :param fill:
    :return:
    """
    # global report_info

    report["canvas"].saveState()

    report["canvas"].setDash(line_style_dict.get(line_style, (1)))
    report["canvas"].setLineWidth(line_width)
    if fore_color is not None:
        report["canvas"].setStrokeColor(colors.HexColor(fore_color))
    if background_color is not None:
        report["canvas"].setFillColor(colors.HexColor(background_color))
    # ellipse(cx, cy, rx, ry, stroke=1, fill=0)
    report["canvas"].ellipse(
        attributes["x"],
        report["cur_y"] - attributes["y"],
        attributes["x"] + attributes["width"],
        report["cur_y"] - attributes["y"] - attributes["height"],
        fill=fill,
    )
    report["canvas"].restoreState()


def process_ellipse(report, element):
    """
    Process jrxml 'ellipse' element
    :param report:.
    :param element:
    """
    # global report_info

    ellipse_element = element.get("child")
    if ellipse_element is not None:
        report_element = process_report_element(
            report, ellipse_element[0].get("reportElement")
        )  # get reportElement

        # set attributes from following graphicElement
        report_element = process_graphic_element(report, ellipse_element, report_element)

        # set line attributes
        line_style = report_element.get("lineStyle")
        line_width = float(report_element.get("lineWidth", "1.0"))
        fore_color = report_element.get("forecolor")
        background_color = report_element.get("backcolor")

        fill = 0
        if background_color is not None:
            fill = 1

        # draw ellipse
        if (
            line_style == "Double"
        ):  # double line consists of wide solid line with thin white line in between
            draw_ellipse(
                report,
                report_element,
                line_style_dict["Solid"],
                line_width * 1.0,
                fore_color,
                background_color,
                fill,
            )
            draw_ellipse(
                report,
                report_element,
                line_style_dict["Solid"],
                line_width * 0.5,
                "#FFFFFF",
                background_color,
                fill,
            )
        else:
            draw_ellipse(
                report, report_element, line_style, line_width, fore_color, background_color, fill
            )


def draw_circle(report, attributes, line_style, line_width, fore_color, background_color, fill):
    """
    Draw an circle.
    :param report:
    :param attributes:
    :param line_style:
    :param line_width:
    :param fore_color:
    :param background_color:
    :param fill:
    :return:
    """
    # global report_info

    report["canvas"].saveState()

    report["canvas"].setDash(line_style_dict.get(line_style, (1)))
    report["canvas"].setLineWidth(line_width)
    if fore_color is not None:
        report["canvas"].setStrokeColor(colors.HexColor(fore_color))
    if background_color is not None:
        report["canvas"].setFillColor(colors.HexColor(background_color))
    # circle(cx, cy, r, stroke=1, fill=0)
    radius = min(attributes["width"], attributes["height"]) / 2
    report["canvas"].circle(
        attributes["x"] + radius, report["cur_y"] - attributes["y"] - radius, radius, fill=fill
    )
    report["canvas"].restoreState()


def process_circle(report, element):
    """
    Process jrxml 'ellipse' element.
    :param element:
    """
    # global report_info

    ellipse_element = element.get("child")
    if ellipse_element is not None:
        report_element = process_report_element(
            report, ellipse_element[0].get("reportElement")
        )  # get reportElement

        # set attributes from following graphicElement
        report_element = process_graphic_element(report, ellipse_element, report_element)

        # set line attributes
        line_style = report_element.get("lineStyle")
        line_width = float(report_element.get("lineWidth", "1.0"))
        fore_color = report_element.get("forecolor")
        background_color = report_element.get("backcolor")

        fill = 0
        if background_color is not None:
            fill = 1

        # draw ellipse
        if (
            line_style == "Double"
        ):  # double line consists of wide solid line with thin white line in between
            draw_circle(
                report,
                report_element,
                line_style_dict["Solid"],
                line_width * 1.0,
                fore_color,
                background_color,
                fill,
            )
            draw_circle(
                report,
                report_element,
                line_style_dict["Solid"],
                line_width * 0.5,
                "#FFFFFF",
                background_color,
                fill,
            )
        else:
            draw_circle(
                report, report_element, line_style, line_width, fore_color, background_color, fill
            )
